import networkx as nx

def bellman_ford(G: nx.Graph,source_node,end_node = None):
    """A brute-force algorithm created via dynamic programming to find
    the shortest path from a source node to all other reachable nodes. It can detect negative edge weight cycles"""
    #Initialize all nodes
    for node in G.nodes():
        G.nodes[node]["parent"] = None
        G.nodes[node]["distance"] = 9999
    G.nodes[source_node]["distance"] = 0

    #Repeatedly relax edges using brute-force + dynamic programming
    for i in range(1,G.number_of_nodes()-1):
        for u,v,w in G.edges.data("weight"):
            weight_plus_distance = G.nodes[u]["distance"] + w
            if weight_plus_distance < G.nodes[v]["distance"]:
                G.nodes[v]["distance"] = weight_plus_distance
                G.nodes[v]["parent"] = u
    
    #Do one last check. Since we've already brute-forced the graph, there shouldn't be any shorter paths UNLESS
    #the graph contains a negative weight cycle (since the shortest path keeps decreasing, forever)
    for u,v,w in G.edges.data("weight"):
        if G.nodes[v]["distance"] > G.nodes[u]["distance"] + w:
            return("Error Contains Negative Weight Cycle")
        
    if end_node != None:
        return backtrack_path(G,source_node,end_node)

    return G.nodes.data("distance")

def backtrack_path(G: nx.Graph,source,end):
    path = []
    i = end
    path.append(i)
    while i != source:
        i = G.nodes[i]["parent"]
        path.append(i)
    path.reverse()
    return (path, G.nodes[end]["distance"])

